# Copyright (c) 2016  Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.

from pygcam.log import getLogger
from .McsSubcommandABC import McsSubcommandABC

_logger = getLogger(__name__)

def driver(args, tool):
    """
    Analyze MCS results
    """
    import os
    from ..analysis import analyzeSimulation
    from ..error import PygcamMcsUserError

    if args.timeseries:
        import pandas as pd
        from pygcam.config import getParam
        from ..Database import getDatabase
        from ..timeseriesPlot import plotTimeSeries, plotForcingSubplots
        from ..util import stripYearPrefix

        simId = args.simId
        expList = args.expName.split(',')
        resultName = args.resultName
        xlabel = 'Year' # args.xlabel or 'Year'
        ymin = args.ymin
        ymax = args.ymax

        # special purpose plot
        forcingPlot = args.forcingPlot

        if not (expList[0] and resultName):
            raise PygcamMcsUserError("expName and resultName must be specified")

        db = getDatabase()
        trialCount = db.getTrialCount(simId)

        plotDir  = getParam('MCS.PlotDir')
        plotType = getParam('MCS.PlotType')

        allResults = db.getTimeSeries(simId, resultName, expList) # , regionName)
        if not allResults:
            raise PygcamMcsUserError('No timeseries results for simId=%d, expList=%s, resultName=%s' \
                                     % (simId, expList, resultName))

        def computeFilename(expName):
            basename = "%s-s%d-%s.%s" % (resultName, simId, expName, plotType)
            filename = os.path.join(plotDir, 's%d' % simId, basename)
            return filename

        # massage the data into the format required by plotTimeSeries
        def createRecord(pair):
            obj, expName = pair
            d = obj.__dict__
            d['expName'] = expName
            return d

        records = [createRecord(pair) for pair in allResults]
        resultDF = pd.DataFrame.from_records(records, index='seriesId')
        units = resultDF.units.iloc[0]
        resultDF.drop(['units', '_sa_instance_state', 'outputId'], axis=1, inplace=True)

        # convert column names like 'y2020' to '2020'
        cols = map(stripYearPrefix, resultDF.columns)
        resultDF.columns = cols

        if forcingPlot:
            filename = computeFilename('combo')
            plotForcingSubplots(resultDF, filename=filename, ci=95, show_figure=False)
            return

        for expName in expList:
            # create a copy so we can drop expName column for melt
            df = resultDF.query("expName == '%s'" % expName).copy()
            _logger.debug("Found %d result records for exp %s, result %s" % (len(df), expName, resultName))

            df.drop(['expName'], axis=1, inplace=True)
            df = pd.melt(df, id_vars=['runId'], var_name='year')

            title = '%s for %s' % (resultName, expName)
            filename = computeFilename(expName)
            _logger.debug('Saving timeseries plot to %s' % filename)

            reg = "" # '-' + regionName if regionName else ""
            extra = "name=%s trials=%d/%d simId=%d scenario=%s%s" % \
                    (resultName, resultDF.shape[0], trialCount, simId, expName, reg)

            # TBD: generalize this with a lookup table or file
            if units == 'W/m^2':
                units = 'W m$^{-2}$'

            plotTimeSeries(df, 'year', 'runId', title=title, xlabel=xlabel, ylabel=units, ci=[95], # [50, 95],
                           text_label=None, legend_name=None, legend_labels=None,
                           ymin=ymin, ymax=ymax, filename=filename, show_figure=False, extra=extra)

        # if doing timeseries plot, none of the other options are relevant
        return

    if not (args.exportInputs or args.resultFile or args.plot or args.importance or
            args.groups or args.plotInputs or args.stats or args.convergence or
            args.exportEMA, args.exportAll):
        msg = 'Must specify at least one of: --export, --resultFile, --plot, --importance, --groups, --distros, --stats, --convergence, --exportEMA, --exportAll'
        raise PygcamMcsUserError(msg)

    analyzeSimulation(args)


class AnalyzeCommand(McsSubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Analyze simulation results stored in the database for the given simulation.
            At least one of -c, -d, -i, -g, -p, -t (or the longname equivalent) must be specified.'''}
        super(AnalyzeCommand, self).__init__('analyze', subparsers, kwargs)

    def addArgs(self, parser):
        parser.add_argument('-c', '--convergence', action='store_true', default=False,
                            help='Generate convergence plots for mean, std dev, skewness, and 95%% coverage interval.')

        parser.add_argument('-d', '--distros', dest='plotInputs', action='store_true', default=False,
                            help='Plot frequency distributions for input parameters.')

        parser.add_argument('-e', '--expName', type=str,
                            help='The name of the experiment or scenario to run.')

        parser.add_argument('-E', '--exportAll', type=str, default=None,
                            help='''Export all inputs for which there are results, and all results for the
                            given expName (-e flag) to the indicated file name.''')

        parser.add_argument('--exportEMA', type=str, default=None,
                            help='''Export results to the given .tar.gz file in a format suitable for analysis
                            using the EMA Workbench. The -e (--expName) and -r (--resultName) flags can hold
                            comma-delimited lists of experiments and results, respectively.''')

        parser.add_argument('--forcingPlot', action='store_true',
                            help='''Plot the data in a good format for multiple forcing timeseries plots''')

        parser.add_argument('-g', '--groups', action='store_true',
                            help='Show the uncertainty importance for groups of parameters.')

        parser.add_argument('-i', '--importance', action='store_true', default=False,
                            help='Show the uncertainty importance for each parameter.')

        parser.add_argument('-l', '--limit', type=int, required=False, default=-1,
                            help='Limit the analysis to the given number of results')

        parser.add_argument('-m', '--min', type=float, required=False, default=None,
                            help='''Limit the analysis to values (for the result named with -r) greater
                            than or equal to this value''')

        parser.add_argument('-M', '--max', type=float, required=False, default=None,
                            help='''Limit the analysis to values (for the result named with -r) less
                            than or equal to this value''')

        parser.add_argument('-o', '--exportInputs', type=str, default=None,
                            help='A file into which to export input (trial) data.')

        parser.add_argument('-O', '--resultFile', type=str, default=None,
                            help='''Export all model results to the given file. When used with this option,
                            the -r (--resultName) and -e (--expName) flags can be comma-delimited lists of
                            result names and experiment names (scenarios), respectively. The output file,
                            in CSV format will have a header (and data in the form) "trialNum,value,expName,resultName"''')

        parser.add_argument('-p', '--plot', action='store_true', default=False,
                            help='''Plot a histogram of the frequency distribution for the named model output
                            (-r required).''')

        parser.add_argument('-R', '--regionName', default=None,
                            help='The region to plot timeseries results for')

        parser.add_argument('-r', '--resultName', type=str, default=None,
                            help='The name of the result variable to analyze.')

        parser.add_argument('-s', '--simId', type=int, default=1,
                            help='The id of the simulation')

        parser.add_argument('-S', '--stats', action='store_true', default=False,
                            help='Print mean, median, max, min, std dev, skewness, and 95%% coverage interval.')

        parser.add_argument('-t', '--timeseries', action='store_true',
                            help='Plot a timeseries distribution')

        parser.add_argument('-x', '--xlabel', dest='xlabel', type=str, default=r'g CO$_2$e MJ$^{-1}$',
                            help='Specify a label for the x-axis in the histogram.')

        parser.add_argument('--ymax', type=float, default=None,
                            help='''Set the scale of a figure by indicating the value to show as the
                            maximum Y value. (By default, scale is set according to the data.)''')

        parser.add_argument('--ymin', type=float, default=None,
                            help='''Set the scale of a figure by indicating the value (given as abs(value),
                            but used as -value) to show as the minimum Y value''')

        return parser   # for auto-doc generation


    def run(self, args, tool):
        driver(args, tool)



'''
.. Created on: 2/26/15

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from __future__ import print_function
import os
import re
import subprocess

from .config import getParam, getParamAsBoolean, parse_version_info, pathjoin, unixPath
from .error import ProgramExecutionError, GcamError, GcamSolverError, PygcamException
from .log import getLogger
from .scenarioSetup import createSandbox
from .utils import writeXmldbDriverProperties, getExeDir, pushd
from .windows import IsWindows

_logger = getLogger(__name__)

PROGRAM = os.path.basename(__file__)

def setJavaPath(exeDir):
    '''
    Update the PATH to be able to find the Java dlls.
    Modeled on run-gcam.bat in the GCAM distribution.
    '''
    if not IsWindows:
        return

    javaHome = os.environ.get('JAVA_HOME', None)

    if not javaHome:
        # Use WriteLocalBaseXDB (4.2) or XMLDBDriver (>= 4.3) to print the java.home property
        # of the Java Runtime used to run it. Note if the runtime is not 64-bit it will only
        # print an error.
        with pushd(exeDir):
            versionInfo = parse_version_info()
            if versionInfo > (4, 2):
                classpath = getParam('GCAM.MI.ClassPath')
                command = 'java -cp "%s" XMLDBDriver --print-java-home' % classpath
            else:
                command = 'java WriteLocalBaseXDB'

            try:
                output = subprocess.check_output(str(command), shell=True)
            except Exception as e:
                raise PygcamException("Cannot get java home dir: %s" % e)

        os.environ['JAVA_HOME'] = javaHome = output and output.strip()

        if not javaHome:
            raise PygcamException("JAVA_HOME not set and failed to read java home directory from WriteLocalBaseXDB")

    if not os.path.isdir(javaHome):
        raise PygcamException('Java home (%s) is not a directory' % javaHome)

    # Update the PATH to be able to find the Java dlls
    # SET PATH=%PATH%;%JAVA_HOME%\bin;%JAVA_HOME%\bin\server"
    javaBin = pathjoin(javaHome, 'bin')
    javaBinServer = pathjoin(javaBin, 'server')
    envPath = os.environ.get('PATH', '')
    os.environ['PATH'] = path = envPath + ';' + javaBin + ';' + javaBinServer
    _logger.debug('PATH=%s', path)

    envClasspath = os.environ.get('CLASSPATH', '')
    envClasspath = ".;" + envClasspath if envClasspath else "."

    miClasspath = getParam('GCAM.MI.ClassPath')
    os.environ['CLASSPATH'] = classpath = envClasspath + ';' + javaBinServer + ';' + miClasspath
    _logger.debug('CLASSPATH=%s', classpath)

def _gcamWrapper(args):
    try:
        _logger.debug('Starting gcam with wrapper')
        gcamProc = subprocess.Popen(args, bufsize=0, stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT, close_fds=True)

    except Exception as e:
        msg = 'gcamWrapper failed to run command: {} ({})'.format(' '.join(args), e)
        raise PygcamException(msg)

    modelDidNotSolve = 'Model did not solve'
    pattern = re.compile('(.*(BaseXException|%s).*)' % modelDidNotSolve)

    gcamOut = gcamProc.stdout
    while True:
        line = gcamOut.readline()
        if line == '':
            break

        _logger.info(line.rstrip())          # see if this ends up in worker.log

        match = re.search(pattern, line)
        if match:
            gcamProc.terminate()
            msg = 'GCAM error: ' + match.group(0)
            if match.group(1) == modelDidNotSolve:
                raise GcamSolverError(msg)
            else:
                raise GcamError(msg)

    _logger.debug('gcamWrapper found EOF. Waiting for GCAM to exit...')
    status = gcamProc.wait()
    _logger.debug('gcamWrapper: GCAM exited with status %s', status)
    return status

#
# Comment taken from exe/run-gcam.command:
#
# We need to find where the Java development kit is installed.
# This could be the Apple supplied version which was provided up
# to 1.6 however was dropped subsequently and instead users may
# have an Oracle provided JDK.  The each take slightly different
# approaches to where libraries live and how to reference them so
# we will have to try to detect the appropriate location.
#
def linkToMacJava():
    import subprocess as subp
    from .error import PygcamException

    cmd = '/usr/libexec/java_home'
    javaHome = ''
    try:
        javaHome = subp.check_output(cmd).strip()
    except Exception:
        pass

    if not javaHome:
        raise PygcamException('Could not find Java install location using "%s"' % cmd)

    # If javaHome contains "1.6", use the Apple supplied version of java 1.6
    libPath = 'lib-stub' if '1.6' in javaHome else javaHome + '/jre/lib/server'

    owd = os.getcwd()
    refWorkspace = getParam('GCAM.RefWorkspace')
    os.chdir(refWorkspace)

    try:
        # Create a symlink to satisfy @rpath searches
        linkName = 'libs/java/lib'
        if not os.path.islink(linkName):
            cmd = "ln -s %s %s" % (libPath, linkName)
            status = subp.call(cmd, shell=True)
            if status != 0:
                raise PygcamException('Failed to create link using "%s"' % cmd)
    finally:
        os.chdir(owd)

def runGCAM(scenario, workspace=None, refWorkspace=None, scenariosDir=None, groupDir='',
            configFile=None, forceCreate=False, noRun=False, noWrapper=False):
    """

    :param scenario: (str) the scenario to run
    :param workspace: (str) path to the workspace to run in, or None, in which
       case the model is run in {GCAM.SandboxDir}/{scenario} if scenario is given
       otherwise, the default scenario in the configuration.xml in the GCAM.RefWorkspace
       is run.
    :param refWorkspace: (str) a workspace to copy files from to create the sandbox,
       if the workspace is not given, or doesn't exist.
    :param scenariosDir: (str) the directory in which the config.xml file for the
       given scenario is found. Defaults to GCAM.ScenariosDir, if given, or "."
    :param groupDir: (str) the name of the scenario group if group sub-directories
       are to be used when computing the location of the scenario's config.xml.
    :param configFile: (str) if scenario is not given, the name of a configuration
       file to run. If scenario is given, this parameter is ignored.
    :param forceCreate: (bool) if True, recreate the sandbox even if it already exists.
    :param noRun: (bool) if True, don't run the model, just create the sandbox and
       display the command that would be executed.
    :param noWrapper: (bool) if True, don't run GCAM inside a "wrapper" that reads
        output and kills the model run as soon as an error is detected.
    :return: none
    :raises ProgramExecutionError: if GCAM exits with non-zero status
    """
    import platform

    if platform.system() == 'Darwin':
        linkToMacJava()

    workspace = workspace or (pathjoin(getParam('GCAM.SandboxDir'), scenario)
                                  if scenario else getParam('GCAM.RefWorkspace'))

    if not os.path.lexists(workspace) or forceCreate:
        createSandbox(workspace, srcWorkspace=refWorkspace, forceCreate=forceCreate)

    exeDir = getExeDir(workspace, chdir=True)
    setJavaPath(exeDir)     # required for Windows; a no-op otherwise

    version = parse_version_info()

    # These features didn't exist in version 4.2
    if version > (4, 2) and not (getParamAsBoolean('GCAM.RunQueriesInGCAM') or
                                 getParamAsBoolean('GCAM.InMemoryDatabase')):    # this implies RunQueriesInGCAM
        # Write a "no-op" XMLDBDriver.properties file
        writeXmldbDriverProperties(inMemory=False, outputDir=exeDir)

    if scenario:
        # Translate scenario name into config file path, assuming that for scenario
        # FOO, the configuration file is {scenariosDir}/{groupDir}/FOO/config.xml
        scenariosDir = unixPath(scenariosDir or getParam('GCAM.ScenariosDir') or '.', abspath=True)
        configFile   = pathjoin(scenariosDir, groupDir, scenario, "config.xml")
    else:
        configFile = unixPath(configFile or pathjoin(exeDir, 'configuration.xml'), abspath=True)

    gcamPath = unixPath(getParam('GCAM.Executable'), abspath=True)
    gcamArgs = [gcamPath, '-C%s' % configFile]  # N.B. GCAM (< 4.2) doesn't allow space between -C and filename

    command = ' '.join(gcamArgs)
    if noRun:
        print(command)
    else:
        _logger.info('Running: %s', command)

        noWrapper = IsWindows or noWrapper     # never use the wrapper on Windows
        exitCode = subprocess.call(gcamArgs, shell=False) if noWrapper else _gcamWrapper(gcamArgs)

        if exitCode != 0:
            raise ProgramExecutionError(command, exitCode)


def gcamMain(args):
    runGCAM(args.scenario, workspace=args.workspace, refWorkspace=args.refWorkspace,
            scenariosDir=args.scenariosDir, groupDir=args.groupDir, configFile=args.configFile,
            forceCreate=args.forceCreate, noRun=args.noRun, noWrapper=args.noWrapper)



"""
.. codeauthor:: Rich Plevin <rich@plevin.com>

.. Copyright (c) 2016  Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
"""
#
# Generate template batch files, then launch ipyparallel
# controller and engines using the values in the template.
#
from __future__ import division, print_function
import copy
from six import iteritems
import os
import stat
import sys
from time import sleep
from IPython.paths import locate_profile

import ipyparallel as ipp
from ipyparallel.apps.ipclusterapp import ALREADY_STARTED, ALREADY_STOPPED, NO_CLUSTER

from .context import Context
from .Database import RUN_NEW, RUN_RUNNING, RUN_SUCCEEDED, RUN_QUEUED, RUN_KILLED, ENG_TERMINATE, getDatabase
from .error import IpyparallelError, PygcamMcsSystemError, PygcamMcsUserError
from .util import parseTrialString, createTrialString
from ..config import getParam, getParamAsInt
from ..log import getLogger

# Exit values for Master.processTrials()
CONTINUE = 1
EXIT = 2

_logger = getLogger(__name__)

#
# SLURM
#
_slurmEngineBatchTemplate = """#!/bin/sh
#SBATCH --account={account}
#SBATCH --partition={queue}
#SBATCH --job-name={cluster_id}-engine
#SBATCH --nodes=1
#SBATCH --tasks-per-node={tasks_per_node}
#SBATCH --time={timelimit}
#{engine_args}
export MCS_WALLTIME={timelimit}
srun %s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}" --prolog="{prolog_script}" --epilog="{epilog_script}"
"""
# These attempts accomplished nothing since TBB doesn't use them
# export OMP_NUM_THREADS=5
# export MKL_NUM_THREADS=1
# export MKL_DOMAIN_NUM_THREADS="BLAS=5"

# No need to run the controller on a compute node, but just in case
_slurmControllerBatchTemplate = """#!/bin/sh
#SBATCH --account={account}
#SBATCH --partition={queue}
#SBATCH --job-name={cluster_id}-controller
#SBATCH --ntasks=1
#SBATCH --time={timelimit}
%s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
"""

#
# PBS
#
_pbsEngineBatchTemplate = """#!/bin/sh
#PBS -V
#PBS -N {cluster_id}-engine
#PBS -q {queue}
#PBS -l walltime={timelimit}
{engine_args}
%s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
"""

_pbsControllerBatchTemplate = """#!/bin/sh
#PBS -V
#PBS -N {cluster_id}-controller
#PBS -q {queue}
#PBS -l walltime={timelimit}
%s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
"""

#
# LSF
#
_lsfEngineBatchTemplate = """#!/bin/sh
TBD
"""

_lsfControllerBatchTemplate = """#!/bin/sh
TBD
"""

# TBD: test PBS (where?)
batchTemplates = {'slurm' : {'engine'     : _slurmEngineBatchTemplate,
                             'controller' : _slurmControllerBatchTemplate},

                  'pbs'   : {'engine'     : _pbsEngineBatchTemplate,
                             'controller' : _pbsControllerBatchTemplate},

                  'lsf'   : {'engine'     : _lsfEngineBatchTemplate,
                             'controller' : _lsfControllerBatchTemplate},
                  }

class Master(object):
    def __init__(self, args):
        self.args = args
        self.db = getDatabase(checkInit=False)
        self.client = None
        self.finished = False

        projectName = args.projectName

        # cache run definitions from the database and amend as necessary when creating runs
        for scenario in args.scenarios:
            rows = self.db.getRunInfo(args.simId, scenario, includeSucceededRuns=False)
            _logger.info('Caching info for %d runs of scenario %s', len(rows), scenario)

            for row in rows:
                assert len(row) == 4, 'db.getRunInfo failed to return 4 values'
                runId, simId, trialNum, status = row
                Context(projectName=projectName, runId=runId, simId=simId,
                        trialNum=trialNum, scenario=scenario, status=status)

    def waitForWorkers(self):
        maxTries  = getParamAsInt('IPP.StartupWaitTries')
        seconds   = getParamAsInt('IPP.StartupWaitSecs')
        profile   = self.args.profile
        clusterId = self.args.clusterId
        client = None

        for i in range(1, maxTries+1):
            if client and len(client) > 0:
                return

            if client is None:
                try:
                    # default timeout is 10 seconds
                    self.client = client = ipp.Client(profile=profile, cluster_id=clusterId)

                # except IOError:
                except Exception as e:
                    _logger.debug("Error waiting for workers: %s", e)

                    if i == maxTries - 1:
                        raise

                    _logger.info("Waiting for client (%d/%d)", i, maxTries)
                    sleep(seconds)
                    continue

            if len(client.ids) == 0:
                _logger.info("Waiting for engines (%d/%d)", i, maxTries)
                sleep(seconds)

        if not client or len(client) == 0:
            raise IpyparallelError('Failed to connect to engines')

    def queueTotals(self):
        """
        Return totals for queue status across all engines
        """
        qstatus = self.client.queue_status()

        unassigned = qstatus.pop(u'unassigned')
        totals = dict(queue=0, completed=0, tasks=0, unassigned=unassigned)

        for eid, qs in iteritems(qstatus):
            for key, count in iteritems(qs):
                totals[key] += count

        return totals

    def shutdownIdleEngines(self):
        '''
        Shutdown the engines.
        '''
        engines = self.client.ids
        engineCount = len(engines)

        if engineCount == 0:
            _logger.info("No engines are running")
            return

        qstatus = self.client.queue_status(targets=engines)

        if qstatus.pop(u'unassigned'):
            return

        # Shutdown idle engines if there are no unassigned tasks
        idleEngines = [id for id, stats in iteritems(qstatus) if stats[u'tasks'] + stats[u'queue'] == 0]

        if idleEngines:
            _logger.info('Shutting down %d idle engines', len(idleEngines))
            _logger.debug('Idle: %s', idleEngines)
            self.client.shutdown(targets=idleEngines, block=True)

            maxTries = 5
            seconds  = 1.5

            expectedEngines = engineCount - len(idleEngines)
            for i in range(maxTries):
                if len(self.client.ids) != expectedEngines:
                    sleep(seconds)

            # TBD: handle timeout waiting for engines to stop
            _logger.info("%d engines active", len(self.client.ids))

    def createRuns(self, simId, scenario, trialNums):
        '''
        Create entries in the "run" table for the given simId, trialNums, and scenario.

        :param simId: (int) simulation ID
        :param scenario: (str) scenario name
        :param trialNums: (list of int) trial numbers
        :return: list of Context instances
        '''
        db = self.db
        session = db.Session()

        try:
            exp = db.getExp(scenario)
            if exp is None:
                raise PygcamMcsUserError("Unknown scenario '%s'" % scenario)

            expId    = exp.expId
            baseline = exp.parent

            runs = []
            projectName = self.args.projectName
            groupName   = self.args.groupName

            # Add a run record for each trial in this chunk
            for trialNum in trialNums:
                # Add a record in the "run" table listing this trial as "new"
                # (row for this simid, trialnum and expid is deleted if it exists)
                run = db.createRun(simId, trialNum, expId=expId, status=RUN_NEW, session=session)
                runs.append(run)

            session.commit()

        except Exception:
            raise

        finally:
            db.endSession(session)

        contexts = map(lambda r: Context(projectName=projectName, runId=r.runId,
                                         simId=simId, trialNum=r.trialNum,
                                         scenario=scenario, groupName=groupName,
                                         baseline=baseline, status=r.status), runs)
        return contexts

    def setRunStatuses(self, pairs):
        """
        Process a list of status changes in a single transaction, e.g., when setting
        the status for a long list of runs to "queued".
        """
        db = getDatabase()
        with db.sessionScope() as session:
            for context, status in pairs:
                self.setRunStatus(context, status=status, session=session)

    def setRunStatus(self, context, status=None, session=None):
        """
        Cache the status of this run, and if it has changed, save the new
        status to the database. Some context objects are retrieved from the
        worker tasks, so we lookup the equivalent in our local cache to test
        for whether a change has occurred.
        """
        debug = False   # set to True for extra diagnostic msgs

        if debug:
            _logger.debug('setRunStatus: %s', context)

        status = status or context.status

        cached = Context.getRunInfo(context.runId)
        if cached:
            if debug:
                _logger.debug('setRunStatus: cache hit: %s', cached)

            if cached.status == status:
                if debug:
                    _logger.debug('setRunStatus: no change; returning')
                return
        else:
            if debug:
                _logger.debug('adding context for runId %d to cache', context.runId)
            cached = context.saveRunInfo()

        _logger.info('%s -> %s', cached, status)
        cached.setVars(status=status)
        self.db.setRunStatus(context.runId, status, session=session)

    def _query_completion_status(self, completed=True):
        # 'completed' flag '$ne' None => running, '$eq' None => completed
        op = ('$ne' if completed else '$eq')

        # db_query returns a list of dicts: [{'msg_id': '.....'}, ...]
        recs = self.client.db_query({'completed': {op: None}}, keys=['msg_id'])

        ids = [rec['msg_id'] for rec in recs]
        return ids

    def runningTasks(self):
        return self._query_completion_status(completed=False)

    def completedTasks(self):
        return self._query_completion_status(completed=True)

    def resubmit(self, task, context):
        _logger.info('Resubmitting task %s', context)
        self.client.resubmit(task)
        self.setRunStatus(context, RUN_QUEUED)

    def getResults(self, tasks):
        if not tasks:
            return None

        client = self.client
        results = []

        for task in tasks:
            try:
                # owner (bool [default: True]) - Whether this AsyncResult should own the result.
                # If so, calling ar.get() will remove data from the client's result and metadata
                # cache. There should only be one owner of any given msg_id.
                ar = client.get_result(task, block=False, owner=False)

                try:
                    if not ar.ready():
                        continue

                except KeyError:        # stale message id can trigger KeyError
                    client.purge_results(jobs=task)
                    continue

                chunk = ar.get()
                workerResult = chunk[0]
                context = workerResult.context
                status = context.status

                if status == ENG_TERMINATE:
                    if ar.engine_id is not None:
                        _logger.info("Terminating engine %s: insufficient time remaining", ar.engine_id)
                        client.shutdown(ar.engine_id)
                        sleep(2)
                        self.resubmit(task, context)

                elif status == RUN_KILLED:
                    self.resubmit(task, context)
                    continue

                else:
                    results.append(workerResult)

            except Exception as e:
                # Raised if an engine dies, e.g., walltime expired.
                # With retries=1, should be able to recover from this.
                _logger.warning('getResults: %s', e)

            client.purge_results(jobs=task)

        return results

    def saveResults(self, results):
        '''
        Called on the master to save results to the database that were prepared by the worker.
        '''
        db = getDatabase()
        session = db.Session()

        try:
            # Delete all old values in first transaction
            for result in results:
                context = result.context
                resultsList = result.resultsList

                if resultsList:
                    # Delete any stale results for this runId (i.e., if re-running a given runId)
                    names = [resultDict['paramName'] for resultDict in resultsList]
                    ids = db.getOutputIds(names)
                    db.deleteRunResults(context.runId, outputIds=ids, session=session)

            db.commitWithRetry(session)

            # Add all new values in a second transaction
            for result in results:
                context = result.context
                resultsList = result.resultsList or []
                runId   = context.runId

                self.setRunStatus(context, session=session)

                if context.status != RUN_SUCCEEDED:
                    continue

                for resultDict in resultsList:
                    paramName  = resultDict['paramName']
                    value      = resultDict['value']
                    regionName = resultDict['regionName']

                    # Save the values to the database
                    if resultDict['isScalar']:
                        db.setOutValue(runId, paramName, value, session=session)
                    else:
                        regionId = db.getRegionId(regionName)   # cached; not a DB query
                        units = resultDict['units']
                        db.saveTimeSeries(runId, regionId, paramName, value, units=units, session=session)

            db.commitWithRetry(session)

        except Exception as e:
            session.rollback()
            # TBD: distinguish database save errors from data access errors?
            raise PygcamMcsSystemError("saveResults failed: %s" % e)

        finally:
            db.endSession(session)

    def checkEngines(self):
        from .slurm import Slurm

        engineSleep = 10
        client = self.client

        while True:
            try:
                if len(client) > 0:
                    return True

                slurm = Slurm()
                pending = slurm.jobsInState('pending', jobName='mcs-engine')    # pending engines, not tasks...

                if len(pending):
                    _logger.info('No engines registered; %d workers PENDING', len(pending))
                    sleep(engineSleep)
                    continue

                else:
                    _logger.info("No engines running or pending. Shutting down hub.")
                    client.shutdown(hub=True, block=True)
                    return False

            except ipp.NoEnginesRegistered:
                sleep(engineSleep)  # handled in loop

    def run(self):
        """
        Run the main wait-and-process loop on `ars`, a list of async result instances.
        Takes parameters from arguments passed from runsim plugin.

        :return: none
        """
        args = self.args

        if args.runLocal:
            self.runTrials()
            return

        if args.redoListOnly and args.statuses:
            listTrialsToRedo(self.db, args.simId, args.scenarios, args.statuses)
            return

        self.waitForWorkers()    # wait for engines to spin up

        shutdownWhenIdle = not args.dontShutdownWhenIdle

        ars = self.runTrials()

        pending = self.client.outstanding

        counter = 0         # for occasionally displaying queue status

        while pending:

            if not self.checkEngines():
                return

            state = 'nominal'   # set to 'completed' when some are completed, to set shorter sleep

            # check for status updates
            for ar in ars:
                data = ar.data[0]
                if data:
                    context = data.get('context')
                    if context:
                        self.setRunStatus(context)

            # finished is the set of msg_ids that are complete
            finished = pending.difference(self.client.outstanding)

            # update pending to exclude those that finished
            pending = pending.difference(finished)

            if finished:
                state = 'completed'
                _logger.debug('%d completed tasks', len(finished))

                results = self.getResults(finished)
                if results:
                    self.saveResults(results)
                else:
                    _logger.warning('Purging %d completed tasks with no results (engine died?)', len(finished))
                    self.client.purge_results(jobs=finished)

                # drop finished tasks from list to avoid checking status needlessly
                toDelete = []
                for ar in ars:
                    id = ar.msg_id[0]
                    if id in finished:
                        toDelete.append(ar)
                        finished.remove(id)

                for ar in toDelete:
                    ars.remove(ar)

                if shutdownWhenIdle:
                    self.shutdownIdleEngines()

            if counter % 5 == 0:
                _logger.info("%d engines: %s", len(self.client.ids), self.queueTotals())

            secs = args.waitSecs if state == 'nominal' else 2
            _logger.debug('sleep(%d)', secs)
            sleep(secs)

            # handle the case of initial over-allocation
            if counter == 0 and shutdownWhenIdle:
                self.shutdownIdleEngines()

            counter += 1

        _logger.info("Shutting down hub")
        self.client.shutdown(hub=True, block=True)

    def runTrials(self):
        import worker

        args = vars(self.args)

        # Construct dict of args to pass to worker tasks
        argDict = {}
        for key in ('runLocal', 'noGCAM', 'noBatchQueries', 'noPostProcessor'):
            argDict[key] = args.get(key, False)

        simId       = args['simId']
        statuses    = args['statuses']
        scenarios   = args['scenarios']
        projectName = args['projectName']
        groupName   = args['groupName']
        trialStr    = args['trials']
        runLocal    = args['runLocal']

        asyncResults = []

        view = None if runLocal else self.client.load_balanced_view(retries=2)

        db = getDatabase()
        exps = {e.expName: e.parent for e in db.getExps()}

        def isBaseline(scenario):
            return (not exps.get(scenario))

        def notBaseline(scenario):
            return exps.get(scenario)

        # sort so baselines come first
        baselines = filter(isBaseline,  scenarios)
        policies  = filter(notBaseline, scenarios)
        scenarios = baselines + policies

        baselineARs = {}      # baseline async_result objects keyed by trialnum

        for scenario in scenarios:

            if statuses:
                # Change this to return Run instances?
                # If any of the "redo" options find trials, use these instead of args.trials
                contexts = self.db.getRunsByStatus(simId, scenario, statuses,
                                                   projectName=projectName,
                                                   groupName=groupName)

                if not contexts:
                    _logger.warn("No trials found for simId=%s, scenario=%s with statuses=%s",
                                 simId, scenario, statuses)
                    continue
            else:
                trialCount = self.db.getTrialCount(simId)

                if trialStr:
                    # convert arg string like "4,7,9-12,42" to a list of ints
                    trialList = parseTrialString(trialStr)
                    userTrials = len(trialList)

                    # remove nonsense values and warn user about them
                    trialNums = filter(lambda trial: 0 <= trial < trialCount, trialList)
                    goodTrials = len(trialNums)
                    if goodTrials != userTrials:
                        _logger.warn('Ignoring %d trial numbers that are out of range [0,%d]',
                                     userTrials - goodTrials, trialCount)
                else:
                    # if trials aren't specified, queue all of them
                    trialNums = range(trialCount)

                contexts = self.createRuns(simId, scenario, trialNums)

            statusPairs = []

            for context in contexts:
                try:
                    if runLocal:
                        self.setRunStatus(context, status=RUN_RUNNING)
                        ctx = copy.copy(context)    # use a copy to simulate what happens with remote call...
                        result = worker.runTrial(ctx, argDict)
                        self.saveResults([result])

                    else:
                        if isBaseline(scenario):
                            result = view.map_async(worker.runTrial, [context], [argDict])
                            baselineARs[context.trialNum] = result

                        else:
                            # TBD: check this works when baseline is not also run (is None the right default, or use []?
                            # If we're also running the baseline for this trial, create a dependency
                            with view.temp_flags(after=baselineARs.get(context.trialNum, None)):
                                result = view.map_async(worker.runTrial, [context], [argDict])

                        statusPairs.append((context, RUN_QUEUED))
                        asyncResults.append(result)

                except Exception as e:
                    _logger.error("Exception running 'runTrial': %s", e)

            self.setRunStatuses(statusPairs)

        return asyncResults


def getTrialsToRedo(db, simId, scenario, statuses):

    # 'missing' is not a real status found in the database
    missing = 'missing' in statuses
    if missing:
        statuses.remove('missing')

    trialNums = []

    # If any of the "redo" options find trials, use these instead of args.trials
    tuples = db.getRunsByStatus(simId, scenario, statuses)
    if tuples:
        trialNums += map(lambda tuple: tuple[2], tuples)  # unpack the (runId, simId, trialNum, ...) tuple

    if missing:
        trialNums += db.getMissingTrials(simId, scenario)

    return trialNums

def listTrialsToRedo(db, simId, scenarios, statuses):
    # 'missing' is not a real status found in the database
    missing = 'missing' in statuses
    if missing:
        statuses.remove('missing')

    for scenario in scenarios:
        trialNums = getTrialsToRedo(db, simId, scenario, statuses)

        if trialNums:
            trialStr = createTrialString(trialNums)
            print("%s:\n%s" % (scenario, trialStr))
        else:
            _logger.warn("No trials found for simId=%s, scenario=%s with statuses=%s",
                         simId, scenario, statuses)

def templatePath(scheduler, profile, clusterId, process):
    profileDir = locate_profile(profile)
    basename = '%s_%s_%s.template' % (scheduler, clusterId, process)
    filename = os.path.join(profileDir, basename)
    return filename

def _saveBatchFiles(numTrials, argDict):
    """
    Create and save engine and controller batch files that will
    be used by ipcontroller and ipengine at start-up. The files are
    created in the profile directory, indicated in `argDict` either
    as "profile_dir" or constructed from "profile" by ipython's
    `locate_profile` function. The ipyparallel config files must set
    the batch_template_file to the names of the generated files, which
    are "{scheduler}-{cluster_id}-{process}.template",
    where {process} is "engine" or "controller", depending as appropriate.
    So for SLURM (the default scheduler), and the default cluster id ("mcs"),
    the two generated files are "slurm-mcs-engine.template" and
    "slurm-mcs-controller.template".

    :param numTrials: (int) the number of GCAM trials to run
    :param argDict: (dict) values for "scheduler", "account", "queue",
       "profile", "cluster_id", "tasks_per_node", "num_engines", "timelimit",
       and "minimum_seconds" are substituted into the batch templates. For
       most of these, defaults are taken from ~/.pygcam.cfg and overridden
       by argDict. Both "ntasks" and "timelimit" are computed on-the-fly.
    :return: (dict) the names of the generated files, keyed by 'engine' or 'controller'
    """
    import pipes

    minutesPerRun = argDict['minutesPerRun']
    maxEngines    = argDict['maxEngines']
    numEngines    = min(numTrials, maxEngines)
    maxRunsPerEngine = numTrials // numEngines + (1 if numTrials % numEngines else 0)

    minutesPerEngine = minutesPerRun * maxRunsPerEngine
    timelimit = "%02d:%02d:00" % (minutesPerEngine // 60, minutesPerEngine % 60)

    defaults = {'scheduler'       : getParam('IPP.Scheduler'),
                'account'         : getParam('IPP.Account'),
                'queue'           : getParam('IPP.Queue'),
                'engine_args'     : getParam('IPP.OtherEngineArgs'),
                'cluster_id'      : argDict['clusterId'],
                'tasks_per_node'  : getParamAsInt('IPP.TasksPerNode'),
                'min_secs_to_run' : getParamAsInt('IPP.MinTimeToRun') * 60,
                'timelimit'       : timelimit,
                'prolog_script'   : getParam('IPP.PrologScript'),
                'epilog_script'   : getParam('IPP.EpilogScript'),
                }

    defaults.update(argDict)

    profile    = defaults['profile']
    scheduler  = defaults['scheduler']
    cluster_id = defaults['cluster_id']

    defaults['profile_dir'] = locate_profile(profile)

    templates = batchTemplates[scheduler.lower()]
    files = {}

    # N.B. process is 'controller' or 'engine'
    for process, template in templates.items():
        cmd_argv = [sys.executable, "-m", "ipyparallel.%s" % process]
        text = template % ' '.join(map(pipes.quote, cmd_argv))  # insert command line
        text = text.format(**defaults)   # insert other parameters

        files[process] = filename = templatePath(scheduler, profile, cluster_id, process)

        _logger.debug('Writing %s', filename)
        with open(filename, 'w') as f:
            f.write(text)

        os.chmod(filename, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)  # make executable

    return files


def _clusterCommand(cmd):
    from pygcam.utils import shellCommand

    status = shellCommand(cmd, shell=True, raiseError=False)

    if status:
        statusStrings = {
            ALREADY_STOPPED: 'Cluster already stopped',
            ALREADY_STARTED: 'Cluster already started',
            NO_CLUSTER:      'No cluster found',
        }
        msg = statusStrings.get(status, 'Exit status: %d' % status)
        _logger.warning(msg)

    return status


def startEngines(numTrials, batchTemplate):
    """
    Uses the batch file created when the cluster was started, so it has
    the profile, cluster-id, and ntasks-per-node already set.
    """
    from pygcam.utils import shellCommand

    tasksPerNode = getParamAsInt('IPP.TasksPerNode')
    maxEngines   = getParamAsInt('IPP.MaxEngines')
    numEngines   = min(numTrials, maxEngines)
    numNodes     = (numEngines // tasksPerNode) + (1 if numEngines % tasksPerNode else 0)

    # TBD: use "ipcluster engines" if there's a way to avoid leaving procs running after shutdown
    # enginesCmd = "ipcluster engines -n %d" % tasksPerNode
    enginesCmd = "sbatch '%s'" % batchTemplate

    for i in range(numNodes):
        _logger.info(enginesCmd)
        status = shellCommand(enginesCmd, shell=True, raiseError=False)
        if status:
            _logger.error('Attempt to launch engines failed; status=%d', status)
            return


def _defaultFromConfig(argDict, name, param):
    '''
    If kwargs doesn't have the named keyword, default to the value of
    the config file param, and set the kwargs dict so this value will
    be used in formatting the template.
    '''
    if argDict.get(name, None) is None:
        argDict[name] = getParam(param)

    return argDict[name]

def pidFileExists(profile, clusterId):
    """
    Check if a process-id (pid) file exists for this profile and clusterId.

    :param profile: (str) profile name
    :param clusterId: (str) cluster ID
    :return: (bool) True if computed pid file pathname exists, else False
    """
    profileDir = locate_profile(profile)
    basename = 'ipcluster-%s.pid' % clusterId
    pidFile = os.path.join(profileDir, 'pid', basename)
    return os.path.isfile(pidFile)


def startCluster(**kwargs):
    """
    Start an ipyparallel cluster to run `numTrials`, using the given `profile`
    and `cluster_id`, if specified, otherwise using values of config vars
    IPP.Profile and IPP.ClusterId, respectively. Writes out a template batch
    file for use with either SLURM or PBS, depending on settings in the user's
    .pygcam.cfg, which is then used by the ipcluster command when launching
    the controller and engines. If the number of trials to run (`numTrials`)
    is greater than the value of config variable IPP.MaxEngines, this max
    number of engines is run but the time limit per engine is increased to allow
    engine each to run the required number of tasks.

    :param clusterId: (str) an id string for the new cluster
    :param numTrials: (int) the number of GCAM trials to run
    :param otherArgs: (str) other arguments appended to the ipcluster command-line
    :param profile: (str) the ipython profile to use
    :param workDir: (str) a directory to change to before starting the cluster
    :return: (int) exit status from the ipcluster command (zero => success)
    """
    # :param maxEngines: (int) the maximum number of engines that will run at once
    # :param minutesPerRun: (int) the time limit to set for each GCAM trial
    # :param queue: (str) the name of the queue to submit the jobs on
    # _defaultFromConfig(kwargs, 'queue',         'IPP.Queue')
    # _defaultFromConfig(kwargs, 'maxEngines',    'IPP.MaxEngines')

    clusterId = _defaultFromConfig(kwargs, 'clusterId', 'IPP.ClusterId')
    profile   = _defaultFromConfig(kwargs, 'profile',   'IPP.Profile')
    _defaultFromConfig(kwargs, 'minutesPerRun', 'IPP.MinutesPerRun')
    _defaultFromConfig(kwargs, 'otherArgs', 'IPP.OtherClusterArgs')
    _defaultFromConfig(kwargs, 'workDir', 'IPP.WorkDir')

    numTrials = kwargs['numTrials']

    if numTrials == 0:
        _logger.warn("numTrials == 0; not starting cluster")
        return 0

    templates = _saveBatchFiles(numTrials, kwargs)
    sleep(1)    # allow files to flush and close

    # If the pid file exists, stop the cluster first
    if pidFileExists(profile, clusterId):
        _logger.info('Stopping cluster "%s"', clusterId)
        stopCluster(profile=profile, cluster_id=clusterId, stop_jobs=True)
        sleep(1)

    # Start cluster with no engines; we add them later based on trial count
    template = "ipcluster start -n 0 --daemonize --profile={profile} --cluster-id={clusterId} --work-dir={workDir} {otherArgs}"
    controllerCmd = template.format(**kwargs)
    status = _clusterCommand(controllerCmd)

    if status == 0:
        startEngines(numTrials, templates['engine'])

    return status


def stopCluster(profile=None, cluster_id=None, stop_jobs=False, other_args=None):
    from pygcam.utils import shellCommand

    # This allows user to pass empty string (e.g., -c='') to override default
    cluster_id = getParam('IPP.ClusterId') if cluster_id is None else cluster_id
    profile    = getParam('IPP.Profile')   if profile    is None else profile
    other_args = getParam('IPP.OtherClusterArgs') if other_args is None else other_args

    template = "ipcluster stop --profile={profile} --cluster-id={cluster_id} {other_args}"
    cmd = template.format(profile=profile, cluster_id=cluster_id, other_args=other_args)

    # shutdown engines, including those that cluster may not be tracking directly
    try:
        client = ipp.Client(profile=profile, cluster_id=cluster_id, timeout=1)
        client.shutdown(hub=False)
    except:
        pass

    status = _clusterCommand(cmd)

    # kill the engines
    if stop_jobs:
        cmd = getParam('IPP.StopJobsCommand').strip()
        if cmd:
            shellCommand(cmd, shell=True, raiseError=False)

    return status


# if __name__ == '__main__':



from .log import getLogger

_logger = getLogger(__name__)

_yearMarketTemplate = '''            <period year="{year}">
              <{marketType} name="{policyName}"/>
            </period>'''

_technologiesTemplate = '''          <technology name="{technology}">
{markets}
          </technology>'''

_policyMarketTemplate = '''<?xml version="1.0" encoding="UTF-8"?>

<!-- This is a generated constraint file. Edits will be overwritten! -->

<scenario>
  <world>
    <region name="{region}">
      <supplysector name="{sector}">
        <subsector name="{subsector}">
{technologies}
        </subsector>
      </supplysector>
    </region>
  </world>
</scenario>
'''

_yearConstraintTemplate = '''        <constraint year="{year}">{level}</constraint>'''

_policyConstraintTemplate = '''<?xml version="1.0" encoding="UTF-8"?>

<!-- This is a generated constraint file. Edits will be overwritten! -->

<scenario>
  <world>
    <region name="{region}">
      <{policyElement} name="{policyName}">
        <policyType>{policyType}</policyType>
        <market>{market}</market>
        {minPrice}
{constraints}
      </{policyElement}>
    </region>
  </world>
</scenario>
'''

DEFAULT_POLICY_ELT  = 'policy-portfolio-standard'
DEFAULT_POLICY_TYPE = 'subsidy'
DEFAULT_MARKET_TYPE = 'input-subsidy'

def policyMarketXml(policyName, region, sector, subsector, technology, years,
                    marketType=DEFAULT_MARKET_TYPE, pathname=None):
    xmlArgs = {}
    xmlArgs['policyName'] = policyName
    xmlArgs['region']     = region
    xmlArgs['sector']     = sector
    xmlArgs['subsector']  = subsector
    xmlArgs['marketType'] = marketType

    techList = technology if isinstance(technology, (list, tuple)) else [technology]
    techNodes = []

    for tech in techList:
        marketList = [_yearMarketTemplate.format(year=year, marketType=marketType, policyName=policyName) for year in years]
        markets = '\n'.join(marketList)
        techNode = _technologiesTemplate.format(markets=markets, technology=tech)
        techNodes.append(techNode)

    xmlArgs['technologies'] = '\n'.join(techNodes)

    xml = _policyMarketTemplate.format(**xmlArgs)
    if pathname:
        _logger.debug("Writing policy market file: {}".format(pathname))
        with open(pathname, 'w') as f:
            f.write(xml)
    else:
        return xml


def policyConstraintsXml(policyName, region, targets, market=None, minPrice=None,
                         policyElement=DEFAULT_POLICY_ELT, policyType=DEFAULT_POLICY_TYPE,
                         pathname=None):
    xmlArgs = {}
    xmlArgs['policyElement'] = policyElement
    xmlArgs['policyType']    = policyType
    xmlArgs['policyName']    = policyName
    xmlArgs['region']        = region
    xmlArgs['market']        = market or region
    xmlArgs['minPrice']      = '<min-price year="1975" fillout="1">{}</min-price>'.format(minPrice) if minPrice != None else ''

    # Generate annual XML for <constraint year="{year}">{level}</constraint>
    constraints = [_yearConstraintTemplate.format(year=year, level=level) for year, level in targets]
    xmlArgs['constraints'] = '\n'.join(constraints)

    xml = _policyConstraintTemplate.format(**xmlArgs)
    if pathname:
        _logger.debug("Writing policy constraint file: {}".format(pathname))
        with open(pathname, 'w') as f:
            f.write(xml)
    else:
        return xml



# Example:
# policyMarketXml('Corn-Etoh-Floor', 'USA', 'ethanol', 'corn ethanol', 'corn ethanol', range(2010, 2051, 10), pathname="/tmp/mkt.xml")
# policyConstraintsXml('Corn-Etoh-Floor', [(2015, 1.6), (2020, 2.0)], 'USA', minPrice=-100, pathname="/tmp/con.xml")
